import 'dart:io';
import 'dart:async';
import 'package:yaml/yaml.dart' as yaml;
import 'package:path/path.dart' as path;
import 'phrase_exceptions.dart';
import 'phrase_arb.dart';
import 'package:glob/glob.dart';
import 'package:build/build.dart';

//inspired by /flutter/packages/flutter_tools/lib/src/localizations/gen_l10n_types.dart
class PhraseGen {
  static const String _TARGET_FILENAME = 'phrase_localizations.dart';
  static const String _TARGET_FILENAME_FULL = '.dart_tool/flutter_gen/gen_l10n/$_TARGET_FILENAME';
  static String _projectFolder = Directory.current.path;

  static Future<String> _getTemplatePath() async {
    var yamlFile = File(path.join(_projectFolder, 'l10n.yaml'));
    if (!await yamlFile.exists()) throw PhraseException("'l10n.yaml' not found.");
    var l10nYaml = yaml.loadYaml(await yamlFile.readAsString());
    var arbDir = l10nYaml['arb-dir'];
    var templateArbFile = l10nYaml['template-arb-file'];
    return path.join(arbDir, templateArbFile);
  }

  static Future<void> generate() async {
    String templatePath = await _getTemplatePath();
    var templateFile = File(path.join(_projectFolder, templatePath));
    var codegen = _generateFromArb(await templateFile.readAsString());
    File genFile = File(_TARGET_FILENAME_FULL);
    await genFile.create(recursive: true);
    await genFile.writeAsString(codegen, mode: FileMode.writeOnly, flush: true);
    print("'${_TARGET_FILENAME}' has been successfully created");
  }

  static String _generateFromArb(String arbRaw){
    var buf = StringBuffer();
    buf.writeln("import 'package:flutter_gen/gen_l10n/app_localizations.dart';");
    buf.writeln("import 'package:flutter_localizations/flutter_localizations.dart';");
    buf.writeln("import 'package:flutter/foundation.dart';");
    buf.writeln("import 'package:flutter/widgets.dart';");
    buf.writeln("import 'package:phrase/phrase.dart';");
    buf.writeln('');
    buf.writeln('class _PhraseLocalizationsDelegate extends LocalizationsDelegate<AppLocalizations> {');
    buf.writeln('\tconst _PhraseLocalizationsDelegate();');
    buf.writeln('\t@override');
    buf.writeln('\tFuture<AppLocalizations> load(Locale locale) => AppLocalizations.delegate.load(locale).then((fallback) => PhraseLocalizations(locale.toString(), fallback)).whenComplete(() => Phrase.onLocale(locale));');
    buf.writeln('\t@override');
    buf.writeln('\tbool isSupported(Locale locale) => AppLocalizations.supportedLocales.contains(locale);');
    buf.writeln('\t@override');
    buf.writeln('\tbool shouldReload(_PhraseLocalizationsDelegate old) => false;');
    buf.writeln('}');
    buf.writeln('');
    buf.writeln('class PhraseLocalizations extends AppLocalizations {');
    buf.writeln('\tPhraseLocalizations(String locale, AppLocalizations fallback) : _fallback = fallback, super(locale);');
    buf.writeln('\tfinal AppLocalizations _fallback;');
    buf.writeln('');
    buf.writeln('\tstatic const LocalizationsDelegate<AppLocalizations> delegate = _PhraseLocalizationsDelegate();');
    buf.writeln('\tstatic const List<LocalizationsDelegate<dynamic>> localizationsDelegates = <LocalizationsDelegate<dynamic>>[');
    buf.writeln('\t\tdelegate,');
    buf.writeln('\t\tGlobalMaterialLocalizations.delegate,');
    buf.writeln('\t\tGlobalCupertinoLocalizations.delegate,');
    buf.writeln('\t\tGlobalWidgetsLocalizations.delegate,');
    buf.writeln('\t];');
    buf.writeln('\tstatic const List<Locale> supportedLocales = AppLocalizations.supportedLocales;');
    buf.writeln('');

    var arb = AppResourceBundle.parse(arbRaw);
    var messages = arb.resourceIds.map((id) => Message(arb.resources, id, false)).toList(growable: false);
    for(var message in messages){
      var id = message.resourceId;
      buf.writeln('\t@override');
      if(message.placeholders.isEmpty){
        buf.writeln("\tString get $id => Phrase.getText(localeName, '$id') ?? _fallback.$id;");
      }else{
        var params = _generateMethodParameters(message).join(', ');
        var values = message.placeholders.map((placeholder) => placeholder.name).join(', ');
        buf.writeln("\tString $id($params) => Phrase.getText(localeName, '$id', [$values]) ?? _fallback.$id($values);");
      }
    }
    buf.writeln('}');
    return buf.toString();
  }

  static List<String> _generateMethodParameters(Message message) {
    assert(message.placeholders.isNotEmpty);
    final countPlaceholder = message.isPlural ? message.getCountPlaceholder() : null;
    return message.placeholders.map((Placeholder placeholder) {
      final type = placeholder == countPlaceholder ? 'num' : placeholder.type;
      return '$type ${placeholder.name}';
    }).toList();
  }
}

Builder phraseBuilder(BuilderOptions options) => PhraseBuilder();

class PhraseBuilder implements Builder {
  @override
  Map<String, List<String>> get buildExtensions => {
    r'$package$': [PhraseGen._TARGET_FILENAME_FULL]
  };

  static final _globArb = new Glob('lib/**.arb');

  @override
  Future<void> build(BuildStep buildStep) async {
    try {
      String templatePath = await PhraseGen._getTemplatePath();
      await for (final input in buildStep.findAssets(_globArb)) {
        if(path.equals(input.path, templatePath)){
          var codegen = PhraseGen._generateFromArb(await buildStep.readAsString(input));
          AssetId output = AssetId(buildStep.inputId.package, PhraseGen._TARGET_FILENAME_FULL);
          await buildStep.writeAsString(output, codegen);
          return;
        }
      }
    }catch(e){
      return;
    }
  }
}
